#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""
The module contains functions for regular expressions.
"""

import re

from yaql.language import specs
from yaql.language import yaqltypes

REGEX_TYPE = type(re.compile('.'))


@specs.parameter('pattern', yaqltypes.String())
def regex(pattern, ignore_case=False, multi_line=False, dot_all=False):
    """:yaql:regex

    Returns regular expression object with provided flags. Can be used for
    matching using matches method.

    :signature: regex(pattern,ignoreCase => false, multiLine => false,
        dotAll => false)
    :arg pattern: regular expression pattern to be compiled to regex object
    :argType pattern: string
    :arg ignoreCase: true makes performing case-insensitive matching.
    :argType ignoreCase: boolean
    :arg multiLine: true makes character '^' to match at the beginning of the
        string and at the beginning of each line, the character '$' to match
        at the end of the string and at the end of each line. false means
        '^' to match only at the beginning of the string, '$' only at the end
        of the string.
    :argType multiLine: boolean
    :arg dotAll: true makes the '.' special character to match any character
        (including a newline). false makes '.' to match anything except
        a newline.
    :argType dotAll: boolean
    :returnType: regex object

    .. code::

        yaql> regex("a.c").matches("abc")
        true
        yaql> regex("A.c").matches("abc")
        false
        yaql> regex("A.c", ignoreCase => true).matches("abc")
        true
    """
    flags = re.UNICODE
    if ignore_case:
        flags |= re.IGNORECASE
    if multi_line:
        flags |= re.MULTILINE
    if dot_all:
        flags |= re.DOTALL
    return re.compile(pattern, flags)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.method
def matches(regexp, string):
    """:yaql:matches

    Returns true if string matches regexp.

    :signature: regexp.matches(string)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to find match in
    :argType string: string
    :returnType: boolean

    .. code::

        yaql> regex("a.c").matches("abc")
        true
    """
    return regexp.search(string) is not None


@specs.parameter('string', yaqltypes.String())
@specs.parameter('regexp', yaqltypes.String())
@specs.method
def matches_(string, regexp):
    """:yaql:matches

    Returns true if string matches regexp, false otherwise.

    :signature: string.matches(regexp)
    :receiverArg string: string to find match in
    :argType string: string
    :arg regexp: regex pattern
    :argType regexp: regex object
    :returnType: boolean

    .. code::

        yaql> "abc".matches("a.c")
        true
    """
    return re.search(regexp, string) is not None


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.name('#operator_=~')
def matches_operator_regex(string, regexp):
    """:yaql:operator =~

    Returns true if left matches right, false otherwise.

    :signature: left =~ right
    :arg left: string to find match in
    :argType left: string
    :arg right: regex pattern
    :argType right: regex
    :returnType: boolean

    .. code::

        yaql> "abc" =~ regex("a.c")
        true
    """
    return regexp.search(string) is not None


@specs.parameter('pattern', yaqltypes.String())
@specs.parameter('string', yaqltypes.String())
@specs.name('#operator_=~')
def matches_operator_string(string, pattern):
    """:yaql:operator =~

    Returns true if left matches right, false otherwise.

    :signature: left =~ right
    :arg left: string to find match in
    :argType left: string
    :arg right: regex pattern
    :argType right: string
    :returnType: boolean

    .. code::

        yaql> "abc" =~ "a.c"
        true
    """
    return re.search(pattern, string) is not None


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.name('#operator_!~')
def not_matches_operator_regex(string, regexp):
    """:yaql:operator !~

    Returns true if left doesn't match right, false otherwise.

    :signature: left !~ right
    :arg left: string to find match in
    :argType left: string
    :arg right: regex pattern
    :argType right: regex
    :returnType: boolean

    .. code::

        yaql> "acb" !~ regex("a.c")
        true
        yaql> "abc" !~ regex("a.c")
        false
    """
    return regexp.search(string) is None


@specs.parameter('pattern', yaqltypes.String())
@specs.parameter('string', yaqltypes.String())
@specs.name('#operator_!~')
def not_matches_operator_string(string, pattern):
    """:yaql:operator !~

    Returns true if left doesn't match right, false otherwise.

    :signature: left !~ right
    :arg left: string to find match in
    :argType left: string
    :arg right: regex pattern
    :argType right: regex object
    :returnType: boolean

    .. code::

        yaql> "acb" !~ regex("a.c")
        true
        yaql> "abc" !~ regex("a.c")
        false
    """
    return re.search(pattern, string) is None


def _publish_match(context, match):
    rec = {
        'value': match.group(),
        'start': match.start(0),
        'end': match.end(0)
    }
    context['$1'] = rec
    for i, t in enumerate(match.groups(), 1):
        rec = {
            'value': t,
            'start': match.start(i),
            'end': match.end(i)
        }
        context['$' + str(i + 1)] = rec

    for key, value, in match.groupdict().values():
        rec = {
            'value': value,
            'start': match.start(value),
            'end': match.end(value)
        }
        context['$' + key] = rec


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('selector', yaqltypes.Lambda(with_context=True))
@specs.method
def search(context, regexp, string, selector=None):
    """:yaql:search

    Search substring which matches regexp. Returns selector applied to
    dictionary {"start" => ..., "end" => ..., "value" => ...} where appropriate
    values describe start of substring, its end and itself. By default, if no
    selector is specified, returns only substring.
    null is a return value if there is no substring which matches regexp.

    :signature: regexp.search(string, selector => null)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to find match in
    :argType string: string
    :arg selector: lambda function to be applied to resulted dictionary with
        keys 'start', 'end', 'value'. null by default, which means to return
        only substring.
    :argType selector: lambda
    :returnType: string or selector return type

    .. code::

        yaql> regex("a.c").search("abcabc")
        "abc"
        yaql> regex("a.c").search("cabc", $)
        {
            "start": 1,
            "end": 4,
            "value": "abc"
        }
        yaql> regex("a.c").search("cabc", $.start)
        1
    """
    res = regexp.search(string)
    new_context = context.create_child_context()
    if res is None:
        return None
    if selector is None:
        return res.group()
    _publish_match(new_context, res)
    return selector(new_context)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('selector', yaqltypes.Lambda(with_context=True))
@specs.method
def search_all(context, regexp, string, selector=None):
    """:yaql:searchAll

    Search all substrings which matches regexp. Returns list of applied to
    dictionary {"start" => ..., "end" => ..., "value" => ...} selector, where
    appropriate values describe start of every substring, its end and itself.
    By default, if no selector is specified, returns only list of substrings.

    :signature: regexp.searchAll(string, selector => null)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to find match in
    :argType string: string
    :arg selector: lambda function to be applied to resulted dictionary of
        every substring with keys 'start', 'end', 'value'. null by default,
        which means to return only list of substrings.
    :argType selector: lambda
    :returnType: list

    .. code::

        yaql> regex("a.c").searchAll("abcadc")
        ["abc", "adc"]
        yaql> regex("a.c").searchAll("abcadc", $)
        [
            {
                "start": 0,
                "end": 3,
                "value": "abc"
            },
            {
                "start": 3,
                "end": 6,
                "value": "adc"
            }
        ]

    :name: searchAll
    """
    for res in regexp.finditer(string):
        new_context = context.create_child_context()
        if selector is None:
            yield res.group()
        else:
            _publish_match(new_context, res)
            yield selector(new_context)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('max_split', int)
@specs.method
def split(regexp, string, max_split=0):
    """:yaql:split

    Splits string by regexp matches and returns list of strings.

    :signature: regexp.split(string, maxSplit => 0)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to be splitted
    :argType string: string
    :arg maxSplit: how many first splits to do. 0 by default, which means
        to split by all matches
    :argType maxSplit: integer
    :returnType: list

    .. code::

        yaql> regex("a.").split("abcadc")
        ["", "c", "c"]
        yaql> regex("a.").split("abcadc", maxSplit => 1)
        ["", "cadc"]
    """
    return regexp.split(string, max_split)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('max_split', int)
@specs.method
@specs.name('split')
def split_string(string, regexp, max_split=0):
    """:yaql:split

    Splits string by regexp matches and returns list of strings.

    :signature: string.split(regexp, maxSplit => 0)
    :receiverArg string: string to be splitted
    :argType string: string
    :arg regexp: regex pattern
    :argType regexp: regex object
    :arg maxSplit: how many first splits to do. 0 by default, which means
        to split by all matches
    :argType maxSplit: integer
    :returnType: list

    .. code::

        yaql> "abcadc".split(regex("a."))
        ["", "c", "c"]
        yaql> "abcadc".split(regex("a."), maxSplit => 1)
        ["", "cadc"]
    """
    return regexp.split(string, max_split)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('repl', yaqltypes.String())
@specs.parameter('count', int)
@specs.method
def replace(regexp, string, repl, count=0):
    """:yaql:replace

    Returns the string obtained by replacing the leftmost non-overlapping
    matches of regexp in string by the replacement repl, where the latter is
    only string-type.

    :signature: regexp.replace(string, repl, count => 0)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to make replace in
    :argType string: string
    :arg repl: string to replace matches of regexp
    :argType repl: string
    :arg count: how many first replaces to do. 0 by default, which means
        to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> regex("a.").replace("abcadc", "xx")
        "xxcxxc"
        yaql> regex("a.").replace("abcadc", "xx", count => 1)
        "xxcadc"
    """
    return regexp.sub(repl, string, count)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('repl', yaqltypes.String())
@specs.parameter('count', int)
@specs.method
@specs.name('replace')
def replace_string(string, regexp, repl, count=0):
    """:yaql:replace

    Returns the string obtained by replacing the leftmost non-overlapping
    matches of regexp in string by the replacement repl, where the latter is
    only string-type.

    :signature: string.replace(regexp, repl, count => 0)
    :receiverArg string: string to make replace in
    :argType string: string
    :arg regexp: regex pattern
    :argType regexp: regex object
    :arg repl: string to replace matches of regexp
    :argType repl: string
    :arg count: how many first replaces to do. 0 by default, which means
        to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> "abcadc".replace(regex("a."), "xx")
        "xxcxxc"
    """
    return replace(regexp, string, repl, count)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('repl', yaqltypes.Lambda(with_context=True))
@specs.parameter('count', int)
@specs.method
def replace_by(context, regexp, string, repl, count=0):
    """:yaql:replaceBy

    Returns the string obtained by replacing the leftmost non-overlapping
    matches of regexp in string by repl, where the latter is an expression to
    get replacements by obtained matches.

    :signature: regexp.replaceBy(string, repl, count => 0)
    :receiverArg regexp: regex pattern
    :argType regexp: regex object
    :arg string: string to make replace in
    :argType string: string
    :arg repl: lambda function which returns string to make replacements
        according to input matches
    :argType repl: lambda
    :arg count: how many first replaces to do. 0 by default, which means
        to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> regex("a.c").replaceBy("abcadc", switch($.value = "abc" => xx,
                                                      $.value = "adc" => yy))
        "xxyy"
    """
    def repl_func(match):
        new_context = context.create_child_context()
        _publish_match(context, match)
        return repl(new_context)
    return regexp.sub(repl_func, string, count)


@specs.parameter('regexp', REGEX_TYPE)
@specs.parameter('string', yaqltypes.String())
@specs.parameter('repl', yaqltypes.Lambda(with_context=True))
@specs.parameter('count', int)
@specs.method
@specs.name('replaceBy')
def replace_by_string(context, string, regexp, repl, count=0):
    """:yaql:replaceBy

    Replaces matches of regexp in string with values provided by the
    supplied function.

    :signature: string.replaceBy(regexp, repl, count => 0)
    :receiverArg string: string to make replace in
    :argType string: string
    :arg regexp: regex pattern
    :argType regexp: regex object
    :arg repl: lambda function which returns string to make replacements
        according to input matches
    :argType repl: lambda
    :arg count: how many first replaces to do. 0 by default, which means
        to do all replacements
    :argType count: integer
    :returnType: string

    .. code::

        yaql> "abcadc".replaceBy(regex("a.c"), switch($.value = "abc" => xx,
                                                      $.value = "adc" => yy))
        "xxyy"
    """
    return replace_by(context, regexp, string, repl, count)


@specs.parameter('string', yaqltypes.String())
def escape_regex(string):
    """:yaql:escapeRegex

    Returns string with all the characters except ASCII letters, numbers,
    and '_' escaped.

    :signature: escapeRegex(string)
    :arg string: string to backslash all non-alphanumerics
    :argType string: string
    :returnType: string

    .. code::

        yaql> escapeRegex('a.')
        "a\\."
    """
    return re.escape(string)


def is_regex(value):
    """:yaql:isRegex

    Returns true if value is a regex object.

    :signature: isRegex(value)
    :arg value: string to backslash all non-alphanumerics
    :argType value: any
    :returnType: boolean

    .. code::

        yaql> isRegex(regex("a.c"))
        true
        yaql> isRegex(regex("a.c").matches("abc"))
        false
    """
    return isinstance(value, REGEX_TYPE)


def register(context):
    context.register_function(regex)
    context.register_function(matches)
    context.register_function(matches_)
    context.register_function(matches_operator_string)
    context.register_function(matches_operator_regex)
    context.register_function(not_matches_operator_string)
    context.register_function(not_matches_operator_regex)
    context.register_function(search)
    context.register_function(search_all)
    context.register_function(split)
    context.register_function(split_string)
    context.register_function(replace)
    context.register_function(replace_by)
    context.register_function(replace_string)
    context.register_function(replace_by_string)
    context.register_function(escape_regex)
    context.register_function(is_regex)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from yaql.language import exceptions
from yaql.language import specs
from yaql.language import yaqltypes
import yaql.tests


class TestTypeAggregation(yaql.tests.TestCase):
    def test_not_of_type(self):
        @specs.parameter('arg', yaqltypes.NotOfType(int))
        def foo(arg):
            return True

        self.context.register_function(foo)
        self.assertTrue(self.eval('foo($)', data='abc'))
        self.assertTrue(self.eval('foo($)', data=[1, 2]))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data=123)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data=True)

    def test_chain(self):
        @specs.parameter(
            'arg', yaqltypes.Chain(yaqltypes.NotOfType(bool), int))
        def foo(arg):
            return True

        self.context.register_function(foo)
        self.assertTrue(self.eval('foo($)', data=123))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data=True)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data='abc')

    def test_any_of(self):
        @specs.parameter(
            'arg', yaqltypes.AnyOf(str, yaqltypes.Integer()))
        def foo(arg):
            if isinstance(arg, str):
                return 1
            if isinstance(arg, int):
                return 2

        self.context.register_function(foo)
        self.assertEqual(1, self.eval('foo($)', data='abc'))
        self.assertEqual(2, self.eval('foo($)', data=123))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data=True)
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo($)', data=[1, 2])



#
#    Copyright (c) 2013-2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import json
import optparse
import sys

import yaql
from yaql.cli import cli_functions
import yaql.legacy


def read_data(f, options):
    if options.string:
        if options.array:
            return [line.rstrip('\n') for line in f]
        else:
            return f.read()
    else:
        if options.array:
            return [json.loads(s) for s in f.readlines()]
        else:
            return json.load(f)


def main():
    p = optparse.OptionParser()
    p.add_option('--data', '-d', help="input file")
    p.add_option('--string', '-s', action='store_true',
                 help="input is a string")
    p.add_option('--native', '-n', action='store_true',
                 help="output data in Python native format")
    p.add_option('--array', '-a', action='store_true',
                 help="read input line by line")
    p.add_option('--tokens', '-t', action='store_true', dest='tokens',
                 help="print lexical tokens info")
    p.add_option('--legacy', '-l', action='store_true', dest='legacy',
                 help="enable legacy v0.2 compatibility mode")
    p.add_option('--limit', '--limit-iterators', '-L', type=int,
                 dest='limit_iterators',
                 default=1000, help="limit iterators by the given number of "
                                    "elements (-1 means infinity)")
    p.add_option('--sets-to-lists', '-S', action="store_true",
                 dest='sets_to_lists', default=True,
                 help="convert all sets in results to lists")
    p.add_option('--tuples-to-lists', '-T', action="store_true",
                 dest='tuples_to_lists', default=False,
                 help="convert all tuples in results to lists")
    p.add_option('--iterable-dicts', '-D', action="store_true",
                 dest='iterable_dicts', default=False,
                 help="consider dictionaries to be iterable over their keys")
    p.add_option('--memory-quota', '--mem', '-m', type=int, dest='memory',
                 default=100000, help="change memory usage quota (in bytes) "
                                      "for all data produced by expressions "
                                      "(-1 means infinity)")
    p.add_option('--keyword-operator', '-k', type=str, dest='keyword_operator',
                 default="=>", help="configure keyword/mapping symbol "
                                    "(empty string means disabled)")
    p.add_option('--allow-delegates', '-A', action="store_true",
                 dest='allow_delegates', default=False,
                 help="enable delegate expression parsing")

    options, arguments = p.parse_args()
    if options.data:
        try:
            if options.data == '-':
                data = read_data(sys.stdin, options)
            else:
                with open(options.data) as f:
                    data = read_data(f, options)
        except Exception:
            print('Unable to load data from ' + options.data, file=sys.stderr)
            exit(1)
    else:
        data = None

    engine_options = {
        'yaql.limitIterators': options.limit_iterators,
        'yaql.convertSetsToLists': options.sets_to_lists,
        'yaql.convertTuplesToLists': options.tuples_to_lists,
        'yaql.iterableDicts': options.iterable_dicts,
        'yaql.memoryQuota': options.memory
    }

    if options.legacy:
        factory = yaql.legacy.YaqlFactory(
            allow_delegates=options.allow_delegates
        )
        context = yaql.legacy.create_context()
        context['legacy'] = True
    else:
        factory = yaql.YaqlFactory(
            allow_delegates=options.allow_delegates,
            keyword_operator=options.keyword_operator
        )
        context = yaql.create_context()

    if options.native:
        context['#nativeOutput'] = True

    parser = factory.create(options=engine_options)
    cli_functions.register_in_context(context, parser)

    if len(arguments) > 0:
        for arg in arguments:
            cli_functions.evaluate(arg, parser, data, context)
    elif options.tokens:
        parser('__main(true)').evaluate(data, context)
    else:
        parser('__main(false)').evaluate(data, context)



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import yaql.tests


class TestMath(yaql.tests.TestCase):
    def test_binary_plus_int(self):
        res = self.eval('2 + 3')
        self.assertEqual(5, res)
        self.assertIsInstance(res, int)

    def test_binary_plus_float(self):
        res = self.eval('2 + 3.0')
        self.assertEqual(5, res)
        self.assertIsInstance(res, float)

        res = self.eval('2.3+3.5')
        self.assertEqual(5.8, res)
        self.assertIsInstance(res, float)

    def test_binary_minus_int(self):
        res = self.eval('12 -3')
        self.assertEqual(9, res)
        self.assertIsInstance(res, int)

    def test_binary_minus_float(self):
        res = self.eval('1 - 2.1')
        self.assertEqual(-1.1, res)
        self.assertIsInstance(res, float)

        res = self.eval('123.321- 0.321')
        self.assertEqual(123.0, res)
        self.assertIsInstance(res, float)

    def test_multiplication_int(self):
        res = self.eval('3 * 2')
        self.assertEqual(6, res)
        self.assertIsInstance(res, int)
        self.assertEqual(-6, self.eval('3 * -2'))
        self.assertEqual(6, self.eval('-3 * -2'))

    def test_multiplication_float(self):
        res = self.eval('3.0 * 2.0')
        self.assertEqual(6.0, res)
        self.assertIsInstance(res, float)
        self.assertAlmostEqual(-6.51, self.eval('3.1 * -2.1'))
        self.assertAlmostEqual(6.51, self.eval('-3.1 * -2.1'))

    def test_division(self):
        self.assertEqual(3, self.eval('7 / 2'))
        self.assertEqual(-4, self.eval('7 / -2'))
        self.assertAlmostEqual(2.5, self.eval('5 / 2.0'))
        self.assertAlmostEqual(2.5, self.eval('5.0 / 2'))
        self.assertAlmostEqual(-2.5, self.eval('-5.0 / 2.0'))
        self.assertRaises(ZeroDivisionError, self.eval, '7 / 0')
        self.assertRaises(ZeroDivisionError, self.eval, '7 / -0.0')

    def test_brackets(self):
        self.assertEqual(-4, self.eval('1 - (2) - 3'))
        self.assertEqual(2, self.eval('1 - (2 - 3)'))

    def test_unary_minus(self):
        self.assertEqual(-4, self.eval('-4'))
        self.assertEqual(-12.0, self.eval('-12.0'))
        self.assertEqual(4, self.eval('3--1'))
        self.assertEqual(2, self.eval('3+-1'))
        self.assertAlmostEqual(4.3, self.eval('3.2 - -1.1'))
        self.assertEqual(2, self.eval('-(1-3)'))

    def test_unary_plus(self):
        self.assertEqual(4, self.eval('+4'))
        self.assertEqual(12.0, self.eval('+12.0'))
        self.assertEqual(2, self.eval('3-+1'))
        self.assertEqual(4, self.eval('3++1'))
        self.assertAlmostEqual(2.1, self.eval('3.2 - +1.1'))

    def test_modulo_int(self):
        res = self.eval('9 mod 5')
        self.assertEqual(4, res)
        self.assertIsInstance(res, int)
        self.assertEqual(-1, self.eval('9 mod -5'))

    def test_modulo_float(self):
        res = self.eval('9.0 mod 5')
        self.assertEqual(4.0, res)
        self.assertIsInstance(res, float)

        res = self.eval('9 mod 5.0')
        self.assertEqual(4.0, res)
        self.assertIsInstance(res, float)

        res = self.eval('9.0 mod 5.0')
        self.assertEqual(4.0, res)
        self.assertIsInstance(res, float)

        self.assertAlmostEqual(-1.1, self.eval('9.1 mod -5.1'))

    def test_abs(self):
        self.assertEqual(4, self.eval('abs(-4)'))
        self.assertEqual(4, self.eval('abs(4)'))
        self.assertEqual(4.4, self.eval('abs(-4.4)'))

    def test_gt(self):
        res = self.eval('5 > 3')
        self.assertIsInstance(res, bool)
        self.assertTrue(res)
        self.assertFalse(self.eval('3 > 3'))

    def test_lt(self):
        res = self.eval('3 < 5')
        self.assertIsInstance(res, bool)
        self.assertTrue(res)
        self.assertFalse(self.eval('3 < 3'))
        self.assertTrue(self.eval('2.5 < 3'))

    def test_gte(self):
        res = self.eval('5 >= 3')
        self.assertIsInstance(res, bool)
        self.assertTrue(res)
        self.assertTrue(self.eval('3 >= 3'))
        self.assertTrue(self.eval('3.5 > 3'))
        self.assertFalse(self.eval('2 >= 3'))

    def test_lte(self):
        res = self.eval('3 <= 5')
        self.assertIsInstance(res, bool)
        self.assertTrue(res)
        self.assertTrue(self.eval('3 <= 3'))
        self.assertFalse(self.eval('3 <= 2'))

    def test_eq(self):
        self.assertTrue(self.eval('5 = 5'))
        self.assertTrue(self.eval('1.0 = 1'))
        self.assertFalse(self.eval('5 = 6'))

    def test_neq(self):
        self.assertFalse(self.eval('5 != 5'))
        self.assertFalse(self.eval('0 != 0.0'))
        self.assertTrue(self.eval('5 != 6'))

    def test_zero_division(self):
        self.assertRaises(ZeroDivisionError, self.eval, '0/0')

    def test_random(self):
        self.assertTrue(self.eval('with(random()) -> $ >= 0 and $ < 1'))
        self.assertTrue(self.eval('with(random(2, 5)) -> $ >= 2 and $ <= 5'))

    def test_int(self):
        self.assertEqual(5, self.eval("int('5')"))
        self.assertEqual(5, self.eval('int(5.2)'))
        self.assertEqual(0, self.eval('int(null)'))

    def test_float(self):
        self.assertAlmostEqual(-1.23, self.eval("float('-1.23')"))
        self.assertEqual(0.0, self.eval('float(null)'))

    def test_bitwise_or(self):
        self.assertEqual(3, self.eval('bitwiseOr(1, 3)'))
        self.assertEqual(3, self.eval('bitwiseOr(1, 2)'))

    def test_bitwise_and(self):
        self.assertEqual(1, self.eval('bitwiseAnd(1, 3)'))
        self.assertEqual(0, self.eval('bitwiseAnd(1, 2)'))

    def test_bitwise_xor(self):
        self.assertEqual(2, self.eval('bitwiseXor(1, 3)'))
        self.assertEqual(3, self.eval('bitwiseXor(1, 2)'))

    def test_bitwise_not(self):
        self.assertEqual(-2, self.eval('bitwiseNot(1)'))

    def test_shift_bits_left(self):
        self.assertEqual(32, self.eval('shiftBitsLeft(1, 5)'))

    def test_shift_bits_right(self):
        self.assertEqual(2, self.eval('shiftBitsRight(32, 4)'))
        self.assertEqual(0, self.eval('shiftBitsRight(32, 6)'))

    def test_pow(self):
        self.assertEqual(32, self.eval('pow(2, 5)'))
        self.assertEqual(4, self.eval('pow(2, 5, 7)'))

    def test_sign(self):
        self.assertEqual(1, self.eval('sign(123)'))
        self.assertEqual(-1, self.eval('sign(-123)'))
        self.assertEqual(0, self.eval('sign(0)'))

    def test_round(self):
        self.assertAlmostEqual(2.0, self.eval('round(2.3)'))
        self.assertAlmostEqual(2.3, self.eval('round(2.345, 1)'))

    def test_is_integer(self):
        self.assertTrue(self.eval('isInteger(-2)'))
        self.assertTrue(self.eval('isInteger(2)'))
        self.assertFalse(self.eval('isInteger(2.3)'))
        self.assertFalse(self.eval('isInteger(abc)'))
        self.assertFalse(self.eval('isInteger(true)'))

    def test_is_number(self):
        self.assertTrue(self.eval('isNumber(-2)'))
        self.assertTrue(self.eval('isNumber(2)'))
        self.assertTrue(self.eval('isNumber(2.3)'))
        self.assertFalse(self.eval('isNumber(abc)'))
        self.assertFalse(self.eval('isNumber(true)'))



#    Copyright (c) 2015 Mirantis, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import tempfile

from yaql.cli.cli_functions import load_data
from yaql.language import exceptions
from yaql.language import specs
from yaql.language import yaqltypes
import yaql.tests


class TestMiscellaneous(yaql.tests.TestCase):
    def test_pass_lambda_from_code(self):
        self.assertEqual(
            [],
            list(self.context('where', self.engine, [1, 2, 3])(False))
        )
        self.assertEqual(
            [2, 3],
            list(self.context('where', self.engine, [1, 2, 3])(
                lambda t: t > 1))
        )

    def test_bool_is_not_an_integer(self):
        @specs.parameter('arg', yaqltypes.Integer())
        def foo(arg):
            return arg

        self.context.register_function(foo)
        self.assertEqual(2, self.eval('foo(2)'))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo(true)')

    def test_nullable_collections(self):
        @specs.parameter('arg', yaqltypes.Sequence())
        def foo1(arg):
            return arg is None

        @specs.parameter('arg', yaqltypes.Sequence(nullable=True))
        def foo2(arg):
            return arg is None

        @specs.parameter('arg', yaqltypes.Iterable())
        def bar1(arg):
            return arg is None

        @specs.parameter('arg', yaqltypes.Iterable(nullable=True))
        def bar2(arg):
            return arg is None

        @specs.parameter('arg', yaqltypes.Iterator())
        def baz1(arg):
            return arg is None

        @specs.parameter('arg', yaqltypes.Iterator(nullable=True))
        def baz2(arg):
            return arg is None

        for func in (foo1, foo2, bar1, bar2, baz1, baz2):
            self.context.register_function(func)

        self.assertFalse(self.eval('foo1([1, 2])'))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'foo1(null)')
        self.assertFalse(self.eval('foo2([1, 2])'))
        self.assertTrue(self.eval('foo2(null)'))

        self.assertFalse(self.eval('bar1([1, 2])'))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'bar1(null)')
        self.assertFalse(self.eval('bar2([1, 2])'))
        self.assertTrue(self.eval('bar2(null)'))

        self.assertFalse(self.eval('baz1($)', data=iter([1, 2])))
        self.assertRaises(
            exceptions.NoMatchingFunctionException,
            self.eval, 'baz1(null)')
        self.assertFalse(self.eval('baz2($)', data=iter([1, 2])))
        self.assertTrue(self.eval('baz2(null)'))

    def test_load_data(self):
        context = {}
        self.assertIsNone(load_data('/temporarydir/some_random_filename',
                                    context))
        self.assertEqual(context, {})

        with tempfile.NamedTemporaryFile() as f:
            f.write(b'This is not JSON')
            f.flush()
            self.assertIsNone(load_data(f.name, context))
            self.assertEqual(context, {})

        with tempfile.NamedTemporaryFile() as f:
            f.write(b'{"foo": "bar"}')
            f.flush()
            self.assertIsNone(load_data(f.name, context))
            self.assertEqual(context['$'], {"foo": "bar"})
